#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Gcmd base command class providing common generic functions.

This is a simple command class for use in gterminal. It provides the basic
methods required for use with gterminal, along with some extra utility methods
that most frontends would be happy to include. If you wish to write your own
command class for gterminal, please do, however ensure that it provides the
necessary methods. Using this class as a template might be a good idea.
"""
# Copyright (C) 2008-2010  James Shubin, McGill University
# Written for McGill University by James Shubin <purpleidea@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
#import struct				# for resize
#import fcntl				# for resize
#import termios
import textwrap				# for help messages
from jhelp import columnize

_ = lambda x: x			# add fake gettext function until i fix up i18n
__all__ = ['base_gcmd']


class base_gcmd:
	"""This is a class similar in style to the python `cmd' class. The
	functionality is often similar or identical, however there are some
	tweaks so that it can play nicely with gterminal. Existing classes
	that inherit from the python cmd class should port to this very easily.
	Any command object for gterminal should inherit from this main class.
	"""
	_gterminal = None	# ref to the `self' of a gterminal instance
	_ruler = '='		# multiple chars allowed

	def __init__(self):
		"""Initialize the gcmd class. The run() method is absolutely
		necessary for this class to work, however set_gterminal() is
		strongly suggested for inclusion for a useful gcmd class."""
		pass


	def set_gterminal(self, gterminal):
		"""Set a handle to the main gterminal class. This is used so
		that code inside of the gcmd class (or a child) can use cool
		methods such as setting and clearing messages, etc..."""
		# what things should get passed in? (at the moment, `self' !)
		# we want this cmd obj to be able to do cool stuff like
		# send gterminal messages, clear them, clear the string
		# and do also sorts of other funky stuff... what is the
		# best way to let cmdobj talk to the gterminal instance?
		self._gterminal = gterminal


	def get_gterminal(self):
		"""Get a handle to the main gterminal class. This function
		exists as the official way for a do_xyz() function to talk
		to the gterminal object and any of its contained references."""
		return self._gterminal


	def parseline(self, line):
		"""Parse a submitted line and return a tuple containing the
		three sanitized values: (command, arg, line). Note that arg
		itself is also a tuple equivalent to what an argv would be."""

		line = line.strip()	# clean up the line
		if len(line) == 0:	# if there's nothing left, return now
			return (None, None, line)

		partitioned = line.partition(' ')
		cmd = partitioned[0].strip()
		arg = partitioned[2].strip()
		if len(arg) == 0: argv = ()	# empty tuple
		else: argv = tuple(arg.split())
		return (cmd, argv, line)


	def parsedoc(self, doc):
		"""My docstring parser. Returns a tuple containing (summary,
		description, usage). Test for truth on a value to see if it is
		present."""
		# see: http://docs.python.org/tutorial/controlflow.html#documentation-strings
		# FIXME: this doesn't respect docstring parsing rules exactly

		USAGE_PREFIX = '-- Usage: $0 '	# usage strings start like this
		l = doc.strip().splitlines()
		summary = ' '.join(l[0:1])	# summary is always first !
		usage = None
		if l[1:2] and l[1:2][0].strip().startswith(USAGE_PREFIX):
				usage = l[1:2][0].partition(USAGE_PREFIX)[2]
				description = ' '.join(l[3:])
		else:		description = ' '.join(l[2:])

		# TODO: swap the replace() func for an actual docstring parser
		description = description.replace('\t', '')	# clean up tabs
		return (summary, description, usage)


	def run(self, line):
		"""Run the command and return any output."""
		cmd, arg, line = self.parseline(line)
		if cmd is None:
			return ''	# returning this seems to work just fine

		if hasattr(self, 'do_%s' % cmd):
			func = getattr(self, 'do_%s' % cmd)
			# run the function inside of a try-except so that if an
			# error occurs, the traceback doesn't bork the terminal
			try:
				result = func(arg)
			# TODO: catch specific exceptions by name and print out
			# more specific errors. maybe use gterminal @ messages.
			except Exception, e:
				self._gterminal.log.error(e)	# log the error
				return self._gterminal.tc.render(
				'%s: ${RED}%s${NORMAL}' % (cmd, _('failed'))
				)
			return result	# return the actual result if it worked
		else:
			return self._gterminal.tc.render(
			'%s: ${RED}%s${NORMAL}' % (cmd, _('command not found'))
			)


	def rmprefix(self, name, prefix='do_'):
		"""Return the leftover after a prefix has been removed."""
		if name.startswith(prefix):
			return name[len(prefix):]
		else:
			return False


	def do_help(self, argv=()):
		"""Displays command help.
		-- Usage: $0 [command]

		Displays help text and usage for a command. When called without
		arguments, available commands are listed."""
		h = _('help: ')		# help prompt text
		H = self._gterminal.tc.render('${BLUE}%s${NORMAL}' % h)
		n = dict([(self.rmprefix(x), getattr(self, x).__doc__) for x in
		dir(self) if self.rmprefix(x)])
		if len(argv) == 0:
			result = '%s\n' % H
			# add a separator across the whole screen width
			result += ('=' * self._gterminal.tc.COLS) + '\n'
			result += columnize.columnize(n.keys()).rstrip('\n')

		elif len(argv) > 1:
			# show help usage, since multiple arguments were passed
			result = self.do_help(('help', ))

		elif argv[0] in n.keys():
			(summary, description, usage) = self.parsedoc(n[argv[0]])
			if usage:
				result = '%s%s %s\n' % (H, argv[0], usage)
			else:
				result = '%s%s\n' % (H, argv[0])

			temp = (' ' * len(h)) + ('%s' % summary)
			result += textwrap.fill(temp, self._gterminal.tc.COLS)
			result += '\n'
			r = str(self._ruler)	# ruler string
			result += (r * (self._gterminal.tc.COLS/len(r)))
			if description: result += '\n'
			# textwrap it to the screen width
			temp = description
			result += textwrap.fill(temp, self._gterminal.tc.COLS)
		else:
			# if help doesn't exist for requested arg
			temp = (argv[0], _('help not found'))
			temp = '%s: ${RED}%s${NORMAL}' % temp
			result = self._gterminal.tc.render(temp)

		return result


	def do_history(self, argv=()):
		"""Displays command history.
		-- Usage: $0 [length]

		Lists previously executed commands. Displays the last ten
		commands if no length argument is specified."""
		# TODO: add commands like history: clear, etc...
		last = 10			# default
		try:
			last = int(argv[0])
		except IndexError, e:		# if argv[0] doesn't exist
			pass
		except TypeError, e:		# if argv[0] isn't an int
			pass
		return self._gterminal.history(last=last)


	def do_exit(self, argv=()):
		"""Causes this interface to terminate."""
		self._gterminal.quit()


class extra_gcmd:

	def do_resize(self, argv=()):
		"""Resizes terminal.
		-- Usage: $0 [col x row]

		If your terminal supports it, resize to: col x row characters,
		or to the default size if no args are specified."""
		# thanks to KANEMOTO Shigeru for showing me what was possible
		# http://omake.accense.com/browser/dual/trunk/dual.py?format=txt
		row = -1
		col = -1
		if len(argv) == 0:
			row = 80		# default size
			col = 24
		else:
			# FIXME: i just rejoined, to avoid rewriting arg parser
			temp = ' '.join(argv)	# use temp variable; do it now!
			count = temp.count('x')
			if count == 1:
				temp = temp.split('x')
				temp = ' '.join(temp)

			temp = temp.strip().split(' ')
			temp = [x for x in temp if x != '']	# remove empty
			if len(temp) == 2:
				row = int(temp[0])
				col = int(temp[1])

		if row > 0 and col > 0:
			fd = sys.stdout.fileno()
			# TODO: don't know if these are required in all cases
			#data = (col, row, 0, 0)
			#winsize = struct.pack('HHHH', *data)
			#fcntl.ioctl(fd, termios.TIOCSWINSZ, winsize)

			# xterm specific escape sequences:
			# ESC [ 4 ; height ; width ; t
			#	set screen size in pixels
			# ESC [ 8 ; height ; width ; t
			#	set screen size in characters
			# ESC [ 14 ; height ; width ; t
			#	request screen size in pixels
			# ESC [ 18 ; height ; width ; t
			#	request screen size in characters
			#os.write(fd, '\033[8;%d;%d;t\033[4;%d;%d;t' % (col, row, 0, 0))
			os.write(fd, '\033[8;%d;%d;t' % (col, row))


