#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Gloader class for mix-in style loading of gcmd style classes.

The gloader class provides functions to allow for mixing-in gcmd style classes
so that arbitrary command style modules can be dynamically loaded and unloaded.
"""
# Copyright (C) 2008-2010  James Shubin, McGill University
# Written for McGill University by James Shubin <purpleidea@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
_ = lambda x: x			# add fake gettext function until i fix up i18n
__all__ = ['gloader', 'build_class']


def build_class(functions=[]):
	"""Pass an array of functions that we want to merge into a class."""
	# example function: please note that the `self' argument is present!
	#
	#	def example_function(self, argv=()):
	#		return 'i am an example function'
	#
	# functions can use the self argument to access: self.get_gterminal()

	# empty class on which to build on. the name: `built' might be seen!
	class built:
		pass

	# TODO: verify the signature is: "as -> s" (tuple of str's -> str)
	for x in functions:
		if hasattr(x, '__name__'):		# plain function
			name = x.__name__
			temp = inspect.getargspec(x)
			assert (temp.args == ['self', 'argv']), \
			_('Bad argument signature in: %s.') % name
			assert (temp.defaults == ((),)), \
			_('Bad argument defaults in: %s.') % name
			# TODO: add assert to verify return value (if possible)

		# XXX: maybe we could take all the functions returned from
		# get_dbus_method() and wrap them in a lambda type function
		# with the same args, so that this inspecting happens the same
		# way as with a regular function...
		elif hasattr(x, '_method_name'):	# dbus proxy function
			name = str(x._method_name)
			# TODO: add asserts to verify signature

		elif False:		# TODO: add cases for xml-rpc and soap
			pass

		else:
			continue			# skip this function

		setattr(built, name, x)		# finally, install the function

	return built


class Base(object):
	"""Empty new-style base class. Needed to keep mixins happy."""
	pass


class gloader(Base):
	"""Loader class providing methods for mixing in other gcmd style
	classes."""

	def __init__(self, cls=[]):
		"""Instantiate this class, and load some friends too. Pass a
		list of classes, and NOT previously instantiated classes."""
		for i in cls: self._load(i)	# load some classes initially


	# TODO: add, change or remove this method. 
	def do_loader(self, argv=()):
		"""Temp method so that we can tell if the loader is present."""
		if len(argv) > 0:
			if argv[0] == 'list':
				result = ''
				for i in gloader.__bases__:
					result += '%s\n' % str(i)
				return result
			else:
				return 'Usage!'
		else:
			return 'The loader is loaded.'


	def _load(self, cls, index=None):
		"""Add a class into this object, before the class specified by
		index. If no index is specified, the class gets appended.
		Integer indexes are also supported. No duplicate insertions
		allowed. Returns boolean success."""
		if cls in gloader.__bases__: return False	# no duplicates
		else:
			temp = list(gloader.__bases__)
			# if index is a class, try to find it as a reference
			if (index is not None) and (type(index) is not int):
				try:
					index = temp.index(index)
				except ValueError, e:
					# do an append instead
					index = None

			if index is None: index = len(temp)	# to append
			if type(index) is int: temp.insert(index, cls)
			else: return False	# should've had an int by now
			gloader.__bases__ = tuple(temp)
			return True


	def _unload(self, cls):
		"""Remove a class from this object. Returns boolean success."""
		if cls in gloader.__bases__:
			temp = list(gloader.__bases__)
			temp.remove(cls)
			# don't ever leave with __bases__ empty!
			if len(temp) == 0: temp = (Base,)
			gloader.__bases__ = tuple(temp)
			return True

		else:
			#raise ValueError(_('%s not in list') % cls)
			return False	# cls not in list, can't remove


	def _reset(self):
		"""Unload all loaded classes."""
		# reset it to the default
		gloader.__bases__ = (Base,)

